from typing import Protocol

from turu.core.protocols.cursor import CursorProtocol


class ConnectionProtocol(Protocol):
    def close(self) -> None:
        ...

    def commit(self) -> None:
        ...

    def rollback(self) -> None:
        ...

    def cursor(self) -> CursorProtocol:
        ...



import contextlib
from pathlib import Path
from typing import (
    Generator,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import turu.core.cursor
from turu.core.protocols.cursor import Parameters
from turu.core.record.csv_recorder import CsvRecorder, CsvRecorderOptions
from turu.core.record.recorder_protcol import RecorderProtcol
from typing_extensions import Unpack

GenericCursor = TypeVar("GenericCursor", bound=turu.core.cursor.Cursor)


class _RecordCursor(
    turu.core.cursor.Cursor[turu.core.cursor.GenericRowType, Parameters]
):
    def __init__(
        self,
        recorder: RecorderProtcol,
        cursor: turu.core.cursor.Cursor[turu.core.cursor.GenericRowType, Parameters],
    ):
        self._recorder = recorder
        self._cursor = cursor

    @property
    def rowcount(self) -> int:
        return self._cursor.rowcount

    @property
    def arraysize(self) -> int:
        return self._cursor.arraysize

    @arraysize.setter
    def arraysize(self, size: int) -> None:
        self._cursor.arraysize = size

    def close(self) -> None:
        self._cursor.close()
        self._recorder.close()

    def execute(
        self, operation: str, parameters: Optional[Parameters] = None, /
    ) -> "_RecordCursor[turu.core.cursor.GenericRowType, Parameters]":
        self._cursor = self._cursor.execute(operation, parameters)

        return self

    def executemany(
        self, operation: str, seq_of_parameters: "Sequence[Parameters]", /
    ) -> "_RecordCursor[turu.core.cursor.GenericRowType, Parameters]":
        self._cursor = self._cursor.executemany(operation, seq_of_parameters)

        return self

    def execute_map(
        self,
        row_type: Type[turu.core.cursor.GenericNewRowType],
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> "_RecordCursor[turu.core.cursor.GenericNewRowType, Parameters]":
        self._cursor = cast(
            turu.core.cursor.Cursor,
            self._cursor.execute_map(row_type, operation, parameters),
        )

        return cast(_RecordCursor, self)

    def executemany_map(
        self,
        row_type: Type[turu.core.cursor.GenericNewRowType],
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> "_RecordCursor[turu.core.cursor.GenericNewRowType, Parameters]":
        self._cursor = cast(
            turu.core.cursor.Cursor,
            self._cursor.executemany_map(row_type, operation, seq_of_parameters),
        )

        return cast(_RecordCursor, self)

    def fetchone(self) -> Optional[turu.core.cursor.GenericRowType]:
        row = self._cursor.fetchone()
        if row is not None:
            self._recorder.write_row(row)

        return row

    def fetchmany(
        self, size: Optional[int] = None
    ) -> List[turu.core.cursor.GenericRowType]:
        rows = self._cursor.fetchmany(size)

        for row in rows:
            self._recorder.write_row(row)

        return rows

    def fetchall(self) -> List[turu.core.cursor.GenericRowType]:
        rows = self._cursor.fetchall()

        for row in rows:
            self._recorder.write_row(row)

        return rows

    def __iter__(self) -> "_RecordCursor[turu.core.cursor.GenericRowType, Parameters]":
        return self

    def __next__(self) -> turu.core.cursor.GenericRowType:
        row = next(self._cursor)

        self._recorder.write_row(row)

        return row


@contextlib.contextmanager
def record_as_csv(
    record_filepath: Union[str, Path],
    cursor: GenericCursor,
    *,
    enable: bool = True,
    **options: Unpack[CsvRecorderOptions],
) -> Generator[GenericCursor, None, None]:
    if enable:
        cursor = cast(
            GenericCursor,
            _RecordCursor(
                CsvRecorder(record_filepath, **options),
                cursor,
            ),
        )

    try:
        yield cursor

    finally:
        cursor.close()



from typing import Optional, Sequence, Type, Union, overload

from turu.core.cursor import GenericRowType
from turu.core.mock.exception import (
    TuruMockResponseTypeMismatchError,
    TuruMockStoreDataNotFoundError,
)


class TuruMockStore:
    def __init__(self):
        self._data = []

    @overload
    def inject_response(
        self,
        row_type: None,
        response: Union[Optional[Sequence[GenericRowType]], Exception] = None,
    ):
        ...

    @overload
    def inject_response(
        self,
        row_type: Type[GenericRowType],
        response: Union[Sequence[GenericRowType], Exception],
    ):
        ...

    def inject_response(
        self,
        row_type: Optional[Type[GenericRowType]],
        response: Union[Sequence[GenericRowType], None, Exception] = None,
    ):
        self._data.append((row_type, response))

    def provide_response(
        self,
        row_type: Optional[Type],
    ) -> list:
        if len(self._data) == 0:
            raise TuruMockStoreDataNotFoundError()

        _row_type, _response = self._data.pop()

        if _row_type is not row_type:
            raise TuruMockResponseTypeMismatchError(row_type, _row_type)

        if isinstance(_response, Exception):
            raise _response

        return _response



from typing import Any, Generic, List, Optional, Protocol, Sequence, TypeVar

from typing_extensions import Self

Parameters = TypeVar("Parameters", contravariant=True)


class CursorProtocol(Generic[Parameters], Protocol):
    @property
    def rowcount(self) -> int:
        """
        the number of rows that the last .execute*() produced (for DQL statements like )
        or affected (for DML statements like or ).

        The attribute is -1 in case no .execute*() has been performed
        on the cursor or the rowcount of the last operation is cannot be determined by the interface.
        """
        ...

    @property
    def arraysize(self) -> int:
        ...

    @arraysize.setter
    def arraysize(self, size: int) -> None:
        ...

    def close(self) -> None:
        ...

    def execute(
        self, operation: str, parameters: Optional[Parameters] = None, /
    ) -> Self:
        ...

    def executemany(
        self, operation: str, seq_of_parameters: Sequence[Parameters], /
    ) -> Self:
        ...

    def fetchone(self) -> Optional[Any]:
        ...

    def fetchmany(self, size: Optional[int] = None) -> List[Any]:
        ...

    def fetchall(self) -> List[Any]:
        ...

    def __iter__(self) -> Self:
        ...

    def __next__(self) -> Any:
        ...



class _NotSupportFeature:
    pass


try:
    import pydantic  # noqa: F401

    USE_PYDANTIC = True
    PydanticModel = pydantic.BaseModel

except ImportError:
    USE_PYDANTIC = False

    PydanticModel = _NotSupportFeature



# from typingsql.protocols.connection import ConnectionProtocol


# class Connection(ConnectionProtocol):
#     def __init__(self, connection: ConnectionProtocol):
#         self._connection = connection



import csv
from dataclasses import is_dataclass
from pathlib import Path
from typing import Union, cast

import turu.core.cursor
from turu.core._feature_flags import USE_PYDANTIC, PydanticModel
from turu.core.exception import TuruRowTypeNotSupportedError
from turu.core.record.recorder_protcol import RecorderProtcol
from typing_extensions import NotRequired, TypedDict, Unpack


class CsvRecorderOptions(TypedDict):
    header: NotRequired[bool]
    rowsize: NotRequired[int]


class CsvRecorder(RecorderProtcol):
    def __init__(
        self,
        filename: Union[str, Path],
        **options: Unpack[CsvRecorderOptions],
    ) -> None:
        self._file = Path(filename).open("w")
        self._writer = csv.writer(self._file)
        self._options = options
        self._writed_rowsize = 0

    def write_row(self, row: turu.core.cursor.RowType) -> None:
        if self._writed_rowsize == 0:
            if self._options.get("header", True):
                self._write_header(row)

        if (rowsize := self._options.get("rowsize")) is not None:
            if self._writed_rowsize >= rowsize:
                return

        self._writed_rowsize += 1

        if is_dataclass(row):
            self._writer.writerow(row.__dict__.values())
            return

        if isinstance(row, tuple):
            self._writer.writerow(cast(tuple, row))
            return

        if USE_PYDANTIC:
            if isinstance(row, PydanticModel):
                self._writer.writerow(row.model_dump().values())
                return

        raise TuruRowTypeNotSupportedError(type(row))

    def _write_header(self, row: turu.core.cursor.RowType) -> None:
        if is_dataclass(row):
            self._writer.writerow(row.__dict__.keys())
            return

        if USE_PYDANTIC:
            if isinstance(row, PydanticModel):
                self._writer.writerow(row.model_dump().keys())
                return

        raise TuruRowTypeNotSupportedError(type(row))

    def close(self) -> None:
        if not self._file.closed:
            self._file.close()



from typing import Any, Iterator, List, Optional, Sequence, Type

from turu.core.cursor import (
    Cursor,
    GenericNewRowType,
    GenericRowType,
    Parameters,
)
from turu.core.mock.exception import (
    TuruMockUnexpectedFetchError,
)
from turu.core.mock.store import TuruMockStore
from typing_extensions import Self, override


class MockCursor(Cursor[GenericRowType, Parameters]):
    def __init__(
        self,
        store: TuruMockStore,
        *,
        row_count: Optional[int] = None,
        rows: Optional[Iterator] = None,
        row_type: Optional[Type[GenericRowType]] = None,
    ) -> None:
        self._turu_mock_store = store
        self._rowcount = row_count
        self._turu_mock_rows = rows
        self._row_type = row_type
        self._arraysize = 1

    @property
    def rowcount(self) -> int:
        return self._rowcount or -1

    @property
    def arraysize(self) -> int:
        return self._arraysize

    @arraysize.setter
    def arraysize(self, size: int) -> None:
        self._arraysize = size

    @override
    def execute(
        self, operation: str, parameters: Optional[Parameters] = None, /
    ) -> "MockCursor[Any, Parameters]":
        return self._make_new_cursor(None)

    @override
    def executemany(
        self, operation: str, seq_of_parameters: Sequence[Parameters], /
    ) -> "MockCursor[Any, Parameters]":
        return self._make_new_cursor(None)

    @override
    def execute_map(
        self,
        row_type: Type[GenericNewRowType],
        operation: str,
        parameters: Optional[Parameters] = None,
        /,
    ) -> "MockCursor[GenericNewRowType, Parameters]":
        return self._make_new_cursor(row_type)

    @override
    def executemany_map(
        self,
        row_type: Type[GenericNewRowType],
        operation: str,
        seq_of_parameters: Sequence[Parameters],
        /,
    ) -> "MockCursor[GenericNewRowType, Parameters]":
        return self._make_new_cursor(row_type)

    @override
    def fetchone(self) -> Optional[GenericRowType]:
        if self._turu_mock_rows is None:
            return None

        try:
            return next(self._turu_mock_rows)

        except StopIteration:
            return None

    @override
    def fetchmany(self, size: Optional[int] = None) -> List[GenericRowType]:
        if self._turu_mock_rows is None:
            raise TuruMockUnexpectedFetchError()

        return [
            next(self._turu_mock_rows)
            for _ in range(size if size is not None else self.arraysize)
        ]

    @override
    def fetchall(self) -> List[GenericRowType]:
        if self._turu_mock_rows is None:
            raise TuruMockUnexpectedFetchError()

        return list(self._turu_mock_rows)

    @override
    def __iter__(self) -> Self:
        if self._turu_mock_rows is None:
            raise TuruMockUnexpectedFetchError()
        return self

    @override
    def __next__(self) -> GenericRowType:
        if self._turu_mock_rows is None:
            raise TuruMockUnexpectedFetchError()

        return next(self._turu_mock_rows)

    def _make_new_cursor(self, row_type: Optional[Type]) -> "MockCursor":
        responses = self._turu_mock_store.provide_response(row_type)

        if responses is None:
            return MockCursor(responses, row_type=row_type)
        else:
            return MockCursor(
                self._turu_mock_store,
                row_count=len(responses),
                rows=iter(responses),
                row_type=row_type,
            )



from abc import abstractmethod


class TuruException(Exception):
    """Base class for all Turu exceptions."""


class TuruError(TuruException):
    """Raised when an error occurs in Turu."""

    @property
    @abstractmethod
    def message(self) -> str:
        ...

    def __str__(self) -> str:
        return self.message


class TuruRowTypeMismatchError(TuruError):
    """Raised when a row is not of the expected type."""

    def __init__(self, expected: type, actual: type) -> None:
        self.expected = expected
        self.actual = actual

    @property
    def message(self) -> str:
        return f"Unsupported row type: Expected {self.expected.__name__}, got {self.actual.__name__}."


class TuruRowTypeNotSupportedError(TuruError):
    """Raised when a row type is not supported."""

    def __init__(self, row_type: type) -> None:
        self.row_type = row_type

    @property
    def message(self) -> str:
        return f"Unsupported row type: {self.row_type.__name__}."


class TuruUnexpectedFetchError(TuruError):
    @property
    def message(self) -> str:
        return "Fetch is unexpected. use execute_*() to specify row type."



from typing import Optional

from turu.core.exception import TuruError


class TuruMockStoreDataNotFoundError(TuruError):
    @property
    def message(self) -> str:
        return "Mock data not found"


class TuruMockResponseTypeMismatchError(TuruError):
    def __init__(self, expected: Optional[type], actual: type) -> None:
        self.expected = expected
        self.actual = actual

    @property
    def message(self) -> str:
        return f"Mock response type mismatch: Expected {self.expected.__name__ if self.expected else None}, got {self.actual.__name__}"


class TuruMockFetchOneSizeError(TuruError):
    def __init__(self, size: int) -> None:
        self.size = size

    @property
    def message(self) -> str:
        return f"Mock fetchone size mismatch: Expected 1, got {self.size}"


class TuruMockUnexpectedFetchError(TuruError):
    @property
    def message(self) -> str:
        return "Mock fetch is unexpected. use execute_map() or executemany_map() to specify row type."
